# -*- coding: utf-8 -*-

__author__ = 'leon'


class ElongException(Exception):
    """Base exception class used by :mod:`~pyelong`"""


class ElongAPIError(ElongException):
    """如果返回值的 `code` 不为 `0`，则 API 调用发生异常"""

    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return '<ElongAPIError ("%s", "%s")>' % (self.code, self.message)



# -*- coding: utf-8 -*-

from pyelong.api.base import ApiBase
from pyelong.api.ihotel.order import Order
from pyelong.api.ihotel.detail import Detail

__all__ = 'Ihotel'


class Ihotel(ApiBase):
    _category = ''

    def list(self, **kwargs):
        """国际酒店列表，方法名：ihotel.list

        文档
        ~~~~
            - http://open.elong.com/wiki/Ihotel.list
        """
        return self._request('list', raw=True, **kwargs)

    @property
    def detail(self):
        return Detail(self._client)

    @property
    def order(self):
        return Order(self._client)



# -*- coding: utf-8 -*-

__author__ = 'leon'

import time
import logging

from pyelong.exceptions import ElongAPIError


class retry_on_error(object):
    _codes_could_retry = {
        'I-0004',  # inner system exception
        'I-0005',  # unknown system exception

        # http://open.elong.com/wiki/%E5%B9%B3%E5%8F%B0%E9%94%99%E8%AF%AF%E7%A0%81
        'A105020003',  # 虚拟卡获取超时
        'A105020005',  # 获取虚拟卡系统错误
        'A103010101',  # 系统错误-国际酒店接口连接失败
        'A203010301',  # 超时重试
        'A103010303',  # 酒店系统错误
        'A203010201',  # 超时重试
        'A101010014',  # 系统错误：Redis连接失败
        'A204010001',  # 超时或系统错误，重试失败请联系艺龙
        'A204010002',  # 加载文件成功，请重试
        'A103050101',  # 系统错误-礼品卡接口 common.giftcard 连接失败
        'A103010401',  # 系统错误-接口连接失败

        # http://open.elong.com/wiki/%E5%9B%BD%E5%86%85%E9%85%92%E5%BA%97%E9%94%99%E8%AF%AF%E7%A0%81
        'H099999999',  # 未知错误
        'H100999',  # 系统异常
        'H101093',  # 底层提交订单保存异常,请重试
    }

    def __init__(self, max_retries=5, delay=0.1, backoff=2,
                 retry_api_error=True, logger=None):
        self.max_retries = max_retries
        self.delay = delay
        self.backoff = backoff
        self.retry_api_error = retry_api_error

        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger

    def is_server_error(self, resp):
        return 500 <= resp.status_code <= 599

    def _sleep(self, tries, delay):
        self.logger.warn('retry [%d], delay: %s ms', tries, delay * 1000)
        time.sleep(delay)
        return delay * self.backoff

    def __call__(self, func):
        def decorator(*args, **kwargs):
            delay = self.delay
            tries = 0
            while tries < self.max_retries:
                tries += 1
                try:
                    resp = func(*args, **kwargs)
                except ElongAPIError as e:
                    if not self.retry_api_error or tries == self.max_retries:
                        raise
                    if e.code not in self._codes_could_retry:
                        raise
                    delay = self._sleep(tries, delay)
                    continue

                if not self.is_server_error(resp) or tries == self.max_retries:
                    return resp

                delay = self._sleep(tries, delay)

        return decorator



# -*- coding: utf-8 -*-

__author__ = 'leon'

from Crypto.Cipher import DES

try:
    from Crypto.Util.Padding import pad, unpad
except ImportError:
    from .padding import pad, unpad


def utf8(value):
    """Converts a string argument to a byte string.

    If the argument is already a byte string or None, it is returned unchanged.
    Otherwise it must be a unicode string and is encoded as utf8.

    Taken from `tornado.escape.utf8`.
    """
    if isinstance(value, bytes):
        return value
    if not isinstance(value, unicode):
        raise TypeError('Expected bytes, unicode; got %r' % type(value))
    return value.encode('utf-8')


def des_encrypt(data, key, iv, mode=DES.MODE_CBC):
    cipher = DES.new(key, mode=mode, IV=iv)
    padded_text = pad(utf8(data), cipher.block_size)
    return cipher.encrypt(padded_text).encode('hex')


def des_decrypt(data, key, iv, mode=DES.MODE_CBC):
    cipher = DES.new(key, mode=mode, IV=iv)
    decrypted = cipher.decrypt(data.decode('hex'))
    return unpad(decrypted, cipher.block_size)



# -*- coding: utf-8 -*-

from pyelong.api.base import ApiBase


class Detail(ApiBase):
    _category = 'ihotel'

    def avail(self, **kwargs):
        """国际酒店详情，方法名：ihotel.detail.avail

        文档
        ~~~~
            - http://open.elong.com/wiki/Ihotel.detail

        NOTE
        ~~~~
            - 请求的入出参数和ihotel.detail一样
            - 属于实时数据接口，将不使用缓存
            - 实时接口访问控制频次更严格(每秒5次)
        """
        return self._request('avail', raw=True, **kwargs)

    def __call__(self, **kwargs):
        """国际酒店详情，方法名：ihotel.detail

        文档
        ~~~~
            - http://open.elong.com/wiki/Ihotel.detail

        NOTE
        ~~~~

        本接口是基于缓存的响应速度快，适用于本地缓存数据或快速展示一个粗略的价格，入参有限制：
            - 只能一个房间
            - 房间大人个数是1或者2，不能带小孩
            - 传入 productId 参数，将不受上面的限制并且结果不使用缓存
        """
        use_detail_avail = False
        room_group = kwargs.get('roomGroup', [])
        if room_group:
            if len(room_group) > 1:
                use_detail_avail = True
            else:
                room = room_group[0]
                if room['childAges'] or room['numberOfAdults'] > 2:
                    use_detail_avail = True

        if use_detail_avail:
            return self._request('avail', raw=True, **kwargs)

        return self._request('', raw=True, **kwargs)



# -*- coding: utf-8 -*-

import hashlib
import json
import time
import urllib

import requests
from requests import RequestException
from tornado import gen
from tornado.httpclient import AsyncHTTPClient

from pyelong.api import ApiSpec
from pyelong.exceptions import ElongAPIError, ElongException
from pyelong.response import RequestsResponse, TornadoResponse, _logger
from pyelong.util.retry import retry_on_error


class Request(object):
    def __init__(self, client,
                 host=ApiSpec.host,
                 version=ApiSpec.version,
                 local=ApiSpec.local):
        self.client = client
        self.verify_ssl = self.client.cert is not None

        self.host = host
        self.version = version
        self.local = local

    def do(self, api, params, https, raw=False):
        raise NotImplementedError()

    def prepare(self, api, params, https, raw):
        timestamp = str(int(time.time()))
        data = self.build_data(params, raw)
        scheme = 'https' if https else 'http'
        url = "%s://%s" % (scheme, self.host)
        params = {
            'method': api,
            'user': self.client.user,
            'timestamp': timestamp,
            'data': data,
            'signature': self.signature(data, timestamp),
            'format': 'json'
        }
        return url, params

    def build_data(self, params, raw=False):
        if not raw:
            data = {
                'Version': self.version,
                'Local': self.local,
                'Request': params
            }
        else:
            data = params
        return json.dumps(data, separators=(',', ':'))

    def signature(self, data, timestamp):
        s = self._md5(data + self.client.app_key)
        return self._md5("%s%s%s" % (timestamp, s, self.client.secret_key))

    @staticmethod
    def _md5(data):
        return hashlib.md5(data.encode('utf-8')).hexdigest()

    def check_response(self, resp):
        if not resp.ok and self.client.raise_api_error:
            _logger.error('pyelong calling api failed, url: %s', resp.url)
            raise ElongAPIError(resp.code, resp.error)
        return resp

    def timing(self, api, delta):
        if self.client.statsd_client and \
                hasattr(self.client.statsd_client, 'timing'):
            self.client.statsd_client.timing(api, delta)


class SyncRequest(Request):
    @property
    def session(self):
        if not hasattr(self, '_session') or not self._session:
            self._session = requests.Session()
        return self._session

    @retry_on_error(retry_api_error=True, logger=_logger)
    def do(self, api, params, https, raw=False):
        url, params = self.prepare(api, params, https, raw)
        try:
            result = self.session.get(url=url,
                                      params=params,
                                      verify=self.verify_ssl,
                                      cert=self.client.cert)
        except RequestException as e:
            _logger.exception('pyelong catches RequestException, url: %s,'
                              ' params: %s', url, params)
            raise ElongException('caught RequestException: %s' % e)
        except Exception as e:
            _logger.exception('pyelong catches unknown exception, url: %s, '
                              'params: %s', url, params)
            raise ElongException('caught unknown exception: %s' % e)

        resp = RequestsResponse(result)
        self.timing(api, resp.request_time)
        return self.check_response(resp)


class AsyncRequest(Request):
    @staticmethod
    def _encode_params(data):
        """
        :param dict data: params

        Taken from requests.models.RequestEncodingMixin._encode_params
        """
        result = []
        for k, vs in data.iteritems():
            if isinstance(vs, basestring) or not hasattr(vs, '__iter__'):
                vs = [vs]
            for v in vs:
                if v is not None:
                    result.append(
                        (k.encode('utf-8') if isinstance(k, str) else k,
                         v.encode('utf-8') if isinstance(v, str) else v))
        return urllib.urlencode(result, doseq=True)

    def _prepare_url(self, url, params):
        if url.endswith('/'):
            url = url.strip('/')
        return '%s?%s' % (url, self._encode_params(params))

    @gen.coroutine
    def do(self, api, params, https, raw=False):
        url, params = self.prepare(api, params, https, raw)
        # use the default SimpleAsyncHTTPClient
        resp = yield AsyncHTTPClient().fetch(self._prepare_url(url, params),
                                             validate_cert=self.verify_ssl,
                                             ca_certs=self.client.cert)
        resp = TornadoResponse(resp)
        self.timing(api, resp.request_time)
        raise gen.Return(self.check_response(resp))
